# -*- coding: utf-8 -*-
# 14-8-19
# create by: snower

import time
import logging
from collections import deque
from tornado.ioloop import IOLoop
from tornado.gen import TracebackFuture
from .transport.stream import TStream as BaseTStream

class TStreamPoolClosedError(Exception):
    pass


class TStream(BaseTStream):
    def __init__(self, *args, **kwargs):
        super(TStream, self).__init__(*args, **kwargs)

        self.used_time = time.time()
        self.idle_time = time.time()


class TStreamPool(object):
    def __init__(self, *args, **kwargs):
        self._max_stream = kwargs.pop("max_stream") if "max_stream" in kwargs else 1
        self._idle_seconds = kwargs.pop("idle_seconds") if "idle_seconds" in kwargs else 0
        self._args = args
        self._kwargs = kwargs
        self._streams = deque()
        self._used_streams = {}
        self._stream_count = 0
        self._wait_streams = deque()
        self._check_idle = False
        self._closed = False
        self._close_callback = None
        self._close_future = None

    def set_close_callback(self, callback):
        self._close_callback = callback

    def stream_close_callback(self, stream):
        try:
            del self._used_streams[id(stream)]
            self._stream_count -= 1
        except KeyError:
            try:
                self._streams.remove(stream)
                self._stream_count -= 1
            except ValueError:
                logging.error("close unknown stream %s", stream)

        if self._closed and not self._streams and not self._used_streams:
            if self._close_callback:
                self._close_callback()
            if self._close_future:
                self._close_future.set_result(True)

    def init_stream(self, future):
        stream = TStream(*self._args, **self._kwargs)
        stream.set_close_callback(lambda : self.stream_close_callback(stream))
        open_future = stream.open()
        self._used_streams[id(stream)] = stream
        self._stream_count += 1

        def finish(open_future):
            if open_future._exc_info is not None:
                future.set_exc_info(open_future.exc_info())
            else:
                future.set_result(stream)
        IOLoop.current().add_future(open_future, finish)

        if self._idle_seconds > 0 and not self._check_idle:
            IOLoop.current().add_timeout(time.time() + self._idle_seconds, self.check_idle)
            self._check_idle = True

    def get_stream(self):
        if self._closed:
            raise TStreamPoolClosedError()

        future = TracebackFuture()
        while self._streams:
            stream = self._streams.pop()
            self._used_streams[id(stream)] = stream
            stream.used_time = time.time()
            if not stream.closed():
                future.set_result(stream)
                return future

        if self._stream_count < self._max_stream:
            self.init_stream(future)
        else:
            self._wait_streams.append(future)
        return future

    def release_stream(self, stream):
        release_future = TracebackFuture()
        release_future.set_result(None)

        if stream.closed():
            return release_future

        if not self._wait_streams:
            if self._closed:
                stream.close()
            else:
                try:
                    del self._used_streams[id(stream)]
                    stream.idle_time = time.time()
                    self._streams.append(stream)
                except KeyError:
                    logging.error("release unknown stream %s", stream)
        else:
            future = self._wait_streams.popleft()
            stream.used_time = time.time()
            future.set_result(stream)

            while self._wait_streams and self._streams:
                stream = self._streams.pop()
                self._used_streams[id(stream)] = stream
                stream.used_time = time.time()
                if not stream.closed():
                    future = self._wait_streams.popleft()
                    future.set_result(stream)

        return release_future

    def check_idle(self):
        for stream in list(self._streams):
            if time.time() - stream.idle_time > self._idle_seconds:
                self._streams.remove(stream)
                self._used_streams[id(stream)] = stream
                stream.close()
        if self._idle_seconds > 0 and self._streams and self._used_streams:
            IOLoop.current().add_timeout(time.time() + self._idle_seconds, self.check_idle)
        else:
            self._check_idle = False

    def close(self):
        self._close_future = TracebackFuture()
        self._closed = True

        while self._wait_streams:
            future = self._wait_streams.popleft()
            future.set_exception(TStreamPoolClosedError())

        while self._streams:
            stream = self._streams.popleft()
            self._used_streams[id(stream)] = stream
            stream.close()
        return self._close_future


# -*- coding: utf-8 -*-
#14-6-11
# create by: snower

import sys
import greenlet
from collections import deque
from tornado.concurrent import TracebackFuture
from tornado.ioloop import IOLoop
from tornado.iostream import StreamClosedError
from thrift.transport import TTransport


if sys.version_info[0] >= 3:
    import io
    StringIO = io.BytesIO
else:
    import cStringIO
    StringIO = cStringIO.StringIO


class TIOStreamTransportFactory:
    """Factory transport that builds buffered transports"""

    def getTransport(self, stream):
        return TIOStreamTransport(stream)


class TIOStreamTransport(TTransport.TTransportBase, TTransport.CReadableTransport):
    DEFAULT_BUFFER = 4096

    def __init__(self, stream):
        self._stream = stream
        self._wbuffer = deque()
        self._wbuffer_len = 0
        self._rbuffer = StringIO(b'')
        self._loop = IOLoop.current()

    def open(self):
        try:
            future = self._stream.open()
        except:
            exc_info = sys.exc_info()
            future = TracebackFuture()
            future.set_exc_info(exc_info)
        return future

    def closed(self):
        return self._stream.closed()

    def close(self):
        if not self.closed():
            return self._stream.close()

    def readAll(self, sz):
        try:
            return self.read(sz)
        except StreamClosedError:
            raise EOFError()

    def read(self, sz):
        data = self._rbuffer.read(sz)
        if len(data) >= sz:
            return data
        partialread = data

        if partialread:
            self._stream._read_buffer.appendleft(partialread)
            self._stream._read_buffer_size += len(partialread)

        if sz <= self._stream._read_buffer_size:
            data, data_len = b''.join(self._stream._read_buffer), self._stream._read_buffer_size
            self._stream._read_buffer.clear()
            self._stream._read_buffer_size = 0

            if data_len == sz:
                return data

            self._rbuffer = StringIO(data)
            return self._rbuffer.read(sz)

        child_gr = greenlet.getcurrent()
        main = child_gr.parent
        assert main is not None, "Execut must be running in child greenlet"

        def read_callback(future):
            if future._exc_info is not None:
                return child_gr.throw(future.exception())

            data = future.result()
            if len(data) == sz:
                return child_gr.switch(data)

            self._rbuffer = StringIO(data)
            return child_gr.switch(self._rbuffer.read(sz))

        future = self._stream.read(sz)
        self._loop.add_future(future, read_callback)
        return main.switch()

    def write(self, data):
        self._wbuffer.append(data)
        self._wbuffer_len += len(data)
        if self._wbuffer_len >= self.DEFAULT_BUFFER:
            data = "".join(self._wbuffer)
            self._wbuffer.clear()
            self._wbuffer_len = 0
            self._stream.write(data)

    def flush(self):
        if self._wbuffer_len:
            data = "".join(self._wbuffer)
            self._wbuffer.clear()
            self._wbuffer_len = 0
            future = self._stream.write(data)
        else:
            future = self._stream.write('')
        if future.done():
            return

        child_gr = greenlet.getcurrent()
        main = child_gr.parent
        assert main is not None, "Execut must be running in child greenlet"

        def write_callback(future):
            if future._exc_info is not None:
                return child_gr.throw(future.exception())
            return child_gr.switch(future.result())
        self._loop.add_future(future, write_callback)
        return main.switch()

    @property
    def cstringio_buf(self):
        return self._rbuffer

    def cstringio_refill(self, partialread, reqlen):
        if partialread:
            self._stream._read_buffer.appendleft(partialread)
            self._stream._read_buffer_size += len(partialread)

        if reqlen <= self._stream._read_buffer_size:
            self._rbuffer = StringIO(b''.join(self._stream._read_buffer))
            self._stream._read_buffer.clear()
            self._stream._read_buffer_size = 0
            return self._rbuffer

        child_gr = greenlet.getcurrent()
        main = child_gr.parent
        assert main is not None, "Execut must be running in child greenlet"

        def read_callback(future):
            if future._exc_info is not None:
                return child_gr.throw(future.exception())

            self._rbuffer = StringIO(future.result())
            return child_gr.switch(self._rbuffer)
        future = self._stream.read(reqlen)
        self._loop.add_future(future, read_callback)
        return main.switch()